# flake8: noqa

from __future__ import absolute_import

from .population import Genotype
from .population import Individual
from .population import Item, Population

from .playgrounds import Static
from .mutations import Uniform
from .recombinations import SimpleArithmetic

from .parents import distributions
from .parents import selections

from .replacements import ReplaceWorst

from .fitness import MultipleLinearRegression